from pydantic import BaseSettings, Field

from ..constants import ENV_PREFIX


class BaseAppConfig(BaseSettings):
    """
    Holds all settings for the latz application. These are parsed from
    config files, command line arguments and environment variables.
    """

    search_backends: tuple[str, ...] = Field(
        default=("unsplash",),
        description="Image search backend to use for retrieving images.",
    )

    class Config:
        env_prefix = ENV_PREFIX



"""
Module which holds everything related to making networking requests for this application.
"""
import asyncio
import logging
from collections.abc import Iterable, Callable

import httpx

logger = logging.getLogger(__name__)


def get_async_client() -> httpx.AsyncClient:
    """
    Returns a httpx.Client object to use for making network requests.

    Note: this is currently pretty sparse but includes room to grow and allows us
    to add settings we wish to apply to all network requests for this particular
    application in the future.
    """
    return httpx.AsyncClient()


async def gather_results(get_callables: Iterable[Callable], limit: int = 10) -> tuple:
    """
    Downloads files asynchronously but limits concurrency to `limit`
    """
    sem = asyncio.Semaphore(limit)  # This allows us to limit our concurrency.

    async def _get(get_callable: Callable):
        async with sem:
            try:
                return await get_callable()
            except httpx.HTTPError as exc:
                logger.error(exc)

    tasks = tuple(_get(get_callable) for get_callable in get_callables)
    return await asyncio.gather(*tasks)



from argparse import Namespace
from typing import cast

import rich_click as click
from pydantic import create_model, validator

from .commands import search_command, config_group
from .config import get_app_config, BaseAppConfig
from .constants import CONFIG_FILES
from .exceptions import ConfigError
from .plugins.manager import get_plugin_manager, AppPluginManager

click.rich_click.USE_RICH_MARKUP = True
click.rich_click.USE_MARKDOWN_EMOJI = True


def create_app_config_class(plugin_manager: AppPluginManager) -> type[BaseAppConfig]:
    """Creates the class that we use to create our application configuration object"""
    # We need to dynamically define our validators because we do not know all the of the
    # valid backends until runtime.
    validators = {
        "backend_validator": validator("search_backends", allow_reuse=True)(
            plugin_manager.get_backend_validator_func()
        )
    }

    # Dynamically create our new configuration object based on possible new fields
    # from our registered plugins.
    return cast(
        type[BaseAppConfig],
        create_model(
            "AppConfig",
            **plugin_manager.search_backend_config_fields,
            __validators__=validators,
            __base__=BaseAppConfig
        ),
    )


@click.group("latz")
@click.pass_context
def cli(ctx):
    """
    "latz" is a command line tool for searching images via various image API backends.
    It is largely meant for educational purposes to show how to develop plugin friendly
    Python applications.

    The included commands are "search" for performing actual image searches and "config"
    for setting and displaying configuration variables.
    """
    ctx.ensure_object(Namespace)
    plugin_manager = get_plugin_manager()
    ctx.obj.plugin_manager = plugin_manager

    AppConfig = create_app_config_class(plugin_manager)
    ctx.obj.config_class = AppConfig

    # Creates the actual config object which parses all possible configuration sources
    # listed in `CONFIG_FILES`.
    try:
        ctx.obj.config = get_app_config(CONFIG_FILES, AppConfig)
    except ConfigError as exc:
        raise click.ClickException(str(exc))


cli.add_command(search_command)
cli.add_command(config_group)



from typing import Literal
from urllib.parse import urljoin

from pydantic import BaseModel, Field

from ...image import (
    ImageSearchResult,
)
from .. import hookimpl, SearchBackendHook

#: Name of the plugin that will be referenced in our configuration
PLUGIN_NAME = "placeholder"


class PlaceholderBackendConfig(BaseModel):
    """
    Configuration for the placeholder backend. Currently only supports "bear"
    or "kitten".
    """

    type: Literal["bear", "kitten"] = Field(
        default="kitten", description="The type of placeholder image links to use"
    )


async def search(client, config, query: str) -> tuple[ImageSearchResult, ...]:
    """
    Search function for placeholder backend. It only returns three pre-defined
    search results. This is primarily meant for testing and demonstration purposes.
    """
    placeholder_type = config.search_backend_settings.placeholder.type
    base_url = f"https://place{placeholder_type}.com"
    sizes = ((200, 300), (600, 500), (1000, 800))

    return tuple(
        ImageSearchResult(
            url=urljoin(base_url, f"{width}/{height}"),
            width=width,
            height=height,
            search_backend=PLUGIN_NAME,
        )
        for width, height in sizes
    )


@hookimpl
def search_backend():
    """
    Registers our placeholder search backend
    """
    return SearchBackendHook(
        name=PLUGIN_NAME, search=search, config_fields=PlaceholderBackendConfig()
    )



from __future__ import annotations

from typing import NamedTuple


class ImageSearchResult(NamedTuple):
    """
    Represents an individual search result object. It holds all relevant data
    for a single search result including the image size and URL.
    """
    url: str | None
    width: int | None
    height: int | None
    search_backend: str | None



import asyncio
from collections.abc import Iterable, Callable
from functools import partial
from itertools import chain

import click
from rich.console import Console
from rich.table import Table

from latz import fetch
from latz.image import ImageSearchResult


def display_results(results: Iterable[ImageSearchResult]) -> None:
    """
    Displays the `ImageSearchResultSet` objects as a `rich.table.Table`
    """
    table = Table(title="Search Results")

    table.add_column("#", no_wrap=True)
    table.add_column("Link", style="magenta")
    table.add_column("Backend", justify="right", style="green")

    for idx, result in enumerate(results, start=1):
        table.add_row(str(idx), result.url, result.search_backend)

    console = Console()
    console.print(table)


async def main(search_callables: Iterable[Callable]):
    """
    Main async coroutine that runs all the currently configured search functions
    and prints the output of the query.
    """
    results = await fetch.gather_results(search_callables)

    # merge results sets into single tuple
    results = tuple(chain(*results))

    display_results(results)


@click.command("search")
@click.argument("query")
@click.pass_context
def command(ctx, query: str):
    """
    Command that retrieves an image based on a search term
    """
    client = fetch.get_async_client()

    # We collect all enabled backends here
    search_backends = ctx.obj.plugin_manager.get_configured_search_backends(
        ctx.obj.config
    )

    # We use `partial` to create a generator with callables preconfigured with the
    # necessary arguments (e.g. `client`, `config` and `query`)
    search_callables = (
        partial(backend.search, client, ctx.obj.config, query)
        for backend in search_backends
    )

    # This is the function call that kicks everything off
    asyncio.run(main(search_callables))



"""
Our exception hierarchy:

* LatzError
    - SearchBackendError
    - ConfigError
"""
from __future__ import annotations


class LatzError(Exception):
    """Base exception for all latz errors"""


class SearchBackendError(LatzError):
    """
    Exception used for raising while encountering issues during image API operations
    """

    def __init__(self, message, original: Exception | None = None):
        self.message = message
        self.original = original


class ConfigError(LatzError):
    pass



from collections.abc import Iterable, Awaitable
from typing import NamedTuple, Any
from collections.abc import Callable

import httpx
import pluggy  # type: ignore
from pydantic import BaseModel

from latz.constants import APP_NAME
from latz.image import ImageSearchResult

hookspec = pluggy.HookspecMarker(APP_NAME)
hookimpl = pluggy.HookimplMarker(APP_NAME)


class SearchBackendHook(NamedTuple):
    """
    Holds the metadata and callable for using the image search hook.
    """

    name: str
    """
    Namespace for the image API plugin; this is the value users will use in their
    config files to specify this particular plugin.

    **Example:**

    ```python
    from latz.plugins import SearchBackendHook

    @hookimpl
    def search_backend():
        return SearchBackendHook(
            name="custom",
            ...
        )
    ```

    This would later be referred to as `custom` in the settings file:

    ```json
    {
      "search_backends": ["custom"],
      "search_backend_settings": {
        "custom": {
          "param_one": "value"
        }
      }
    }
    ```
    """

    search: Callable[
        [httpx.AsyncClient, Any, str], Awaitable[tuple[ImageSearchResult, ...]]
    ]
    """
    Callable that implements the search hook.
    """

    config_fields: BaseModel
    """
    Pydantic model that defines all the settings that this plugin needs.

    **Example:**

    ```python

    from pydantic import BaseModel, Field

    PLUGIN_NAME = "custom"

    class CustomConfigFields(BaseModel):
        access_key: str = Field(description="Access key for the API")

    @hookimpl
    def image_api():
        return ImageAPIPlugin(
            name=PLUGIN_NAME,
            config_fields=CustomConfigFields(access_key=""),
            ...
        )
    ```
    """


class AppHookSpecs:
    """Holds all hookspecs for this application"""

    @hookspec
    def search_backend(self) -> Iterable[SearchBackendHook]:
        """
        Hookspec for the search backend hook.

        Check out the [creating plugins][creating-plugins] guide for more information on
        using this plugin hook.
        """
        return tuple()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['latz',
 'latz.commands',
 'latz.commands.config',
 'latz.config',
 'latz.plugins',
 'latz.plugins.image']

package_data = \
{'': ['*']}

modules = \
['py']
install_requires = \
['Pillow>=9.3.0,<10.0.0',
 'click>=8.1.3,<9.0.0',
 'httpx>=0.23.1,<0.24.0',
 'pluggy>=1.0.0,<2.0.0',
 'pydantic>=1.10.7,<2.0.0',
 'rich-click>=1.6.0,<2.0.0',
 'rich>=13.3.3,<14.0.0']

entry_points = \
{'console_scripts': ['latz = latz.cli:cli']}

setup_kwargs = {
    'name': 'latz',
    'version': '0.2.1',
    'description': 'CLI Program for downloading images. Maybe by location too...',
    'long_description': '# Overview\n\n[pluggy]: https://pluggy.readthedocs.io/en/stable/\n[click]: https://click.palletsprojects.com/\n[pydantic]: https://docs.pydantic.dev/\n[rich]: https://rich.readthedocs.io/\n[anaconda.org]: https://anaconda.org\n[latz-imgur]: https://github.com/travishathaway/latz-imgur\n[creating-plugins]: creating-plugins\n\nThis is a command line tool used for retrieving images from various image\nsearch backends (e.g. Unsplash, Google). This tool is primarily developed for educational purposes\nto show people how to develop plugin friendly Python applications. Furthermore,\nit is an example project that shows how to effectively pair a handful of\npopular Python libraries to write command line applications.\n\nTo facilitate our plugin architecture, the [pluggy][pluggy] library is used.\nOther libraries used include the following:\n\n- [click][click]: used for structuring the command line application 🖱 💻\n- [pydantic][pydantic]: used for handling configuration file validation 🗃\n- [rich][rich]: used for UX/UI elements and generally making the application more pretty 🌈\n\n### Why "latz"\n\n"latz" is short and easy to type! This is super important when writing CLI programs.\nI also might add a geolocation search feature, so it is a reference  to the word "latitude".\n\n## Quick Start\n\n### Installation\n\nlatz is available for install either on PyPI:\n\n```bash\n# Run from a new virtual environment\n$ pip install latz\n```\n\nor my own [anaconda.org][anaconda.org] channel:\n\n```bash\n$ conda create -n latz \'thath::latz\'\n```\n\nIf you are interested in tinkering around with the code yourself, you can also\nrun it locally:\n\n```bash\n$ git clone git@github.com:/travishathaway/latz.git\n$ cd latz\n# Create a virtual environment however you like..\n$ pip install -e .\n```\n\n### Usage\n\nLatz comes initially configured with the "unsplash" image search backend. To use this,\nyou will need to create an Unsplash account and create a test application. After getting\nyour "access_key", you can set this value by running this command:\n\n```bash\n$ latz config set search_backend_settings.unsplash.access_key=<YOUR_ACCESS_KEY>\n```\n\nOnce this is configured, you can search Unsplash for bunny pictures:\n\n```bash\n$ latz search "bunny"\n[\n    ImageSearchResultSet(\n        results=(\n            ImageSearchResult(\n                url=\'https://unsplash.com/photos/u_kMWN-BWyU/download?ixid=MnwzOTMwOTR8MHwxfHNlYXJjaHwxfHxidW5ueXxlbnwwfHx8fDE2Nzk0MTA2NzQ\',\n                width=3456,\n                height=5184\n            ),\n            # ... results truncated\n        ),\n        total_number_results=10,\n        search_backend=\'unsplash\'\n    )\n]\n```\n\n### Configuring\n\nThe configuration for latz is stored in your home direct and is in the JSON format.\nBelow is a what a default version of this configuration looks like:\n\n```json\n{\n  "search_backends": [\n    "unsplash"\n  ],\n  "search_backend_settings": {\n    "placeholder": {\n      "type": "kitten"\n    },\n    "unsplash": {\n      "access_key": "your-access-key"\n    }\n  }\n}\n```\n\n_Latz will also search in your current working directory for a `.latz.json` file and use this in your configuration.\nFiles in the current working directory will be prioritized over your home directory location._\n\nTo see other available image search backends, see [Available image search backends](#available-image-search-backends) below.\n\n### Available image search backends\n\nHere are a list of the available search backends:\n\n#### Built-in\n\n- "unsplash"\n- "placeholder"\n\n#### Third-party\n\n- [latz-imgur][latz-imgur]\n\n### How to extend and write your own image search backend\n\nPlease see the [creating plugins][creating-plugins] guide in the documentation.\n',
    'author': 'Travis Hathaway',
    'author_email': 'travis.j.hathaway@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from collections import Counter
from functools import reduce
from collections.abc import Callable

from pluggy import PluginManager  # type: ignore
from pydantic import create_model

from ..constants import APP_NAME
from ..exceptions import LatzError
from .hookspec import AppHookSpecs, SearchBackendHook
from .image import unsplash, placeholder

#: Name of the dynamically generated pydantic model for search backend settings.
#: These settings are registered via plugins.
SEARCH_BACKEND_SETTINGS_MODEL = "SearchBackendSettings"

#: Key used in the configuration of the dynamically generated settings from plugins.
SEARCH_BACKEND_SETTINGS_KEY = "search_backend_settings"


class AppPluginManager(PluginManager):
    """
    Our own custom subclass of PluginManager. We do this to add several convenience methods
    we use elsewhere in our application.
    """

    def __init__(self, *args, **kwargs):
        # This is used as a cache
        self.__search_backend_hooks = None

        super().__init__(*args, **kwargs)

    def reset_cache(self) -> None:
        """Resets all cached properties"""
        self.__search_backend_hooks = None

    @property
    def search_backend_names(self) -> tuple[str, ...]:
        """
        Get the names of available search backends that are currently configured.
        We cache this once per run.

        If there are duplicate names registered, we raise an exception.

        :raises LatzError: Raised if duplicate values are found (same plugin
                                is registered multiple times)
        """
        if self.__search_backend_hooks is not None:
            return self.__search_backend_hooks

        self.__search_backend_hooks = tuple(
            search_backend.name for search_backend in self.hook.search_backend()
        )
        names_counter = Counter(self.__search_backend_hooks)
        duplicates = tuple(value for value, count in names_counter.items() if count > 1)

        if len(duplicates) > 0:
            raise LatzError(
                "Duplicate values for the hook 'search_backend' found: "
                f"{', '.join(duplicates)}. Please make sure to define a unique 'name' field"
            )

        return self.__search_backend_hooks

    def get_configured_search_backends(self, config) -> tuple[SearchBackendHook, ...]:
        """
        Get the search backends that are currently configured to be used.
        These are differently than those that have simply been registered.
        """
        return tuple(
            search_backend
            for search_backend in self.hook.search_backend()
            if search_backend.name in config.search_backends
        )

    @property
    def search_backend_config_fields(self) -> dict:
        """
        Returns all the registered config fields for the `search_backend` plugins.
        We perform a merge of all registered `config_fields` dictionaries that represent
        the new configuration fields that will be added.

        TODO: add a duplicate check just like for search backend names
        """
        # Merge config fields from all registered plugins
        search_backend_config = reduce(
            lambda dict_one, dict_two: {**dict_one, **dict_two},
            (
                {search_backends.name: search_backends.config_fields}
                for search_backends in self.hook.search_backend()
            ),
        )

        SearchBackendSettings = create_model(
            SEARCH_BACKEND_SETTINGS_MODEL, **search_backend_config
        )

        return {
            SEARCH_BACKEND_SETTINGS_KEY: (
                SearchBackendSettings,
                SearchBackendSettings().dict(),
            )
        }

    def get_backend_validator_func(self) -> Callable:
        """Returns the validator function that is used by Pydantic when parsing configuration"""

        def validate_backend(cls, values):
            for value in values:
                if value not in self.search_backend_names:
                    valid_names = ", ".join(self.search_backend_names)
                    raise ValueError(
                        f"'{value}' is not a valid choice for a search backend. "
                        f"Available choices: {valid_names}"
                    )
            return values

        return validate_backend


def get_plugin_manager() -> AppPluginManager:
    """
    Plugin manager for the application. This function registers all plugin
    hooks, internal and external plugins.
    """
    # This is the plugin manager object we return that the application will use
    # to get information about register plugins.
    plugin_manager = AppPluginManager(APP_NAME)

    # Adds plugin hooks
    plugin_manager.add_hookspecs(AppHookSpecs)

    # Registers internal plugin hooks
    plugin_manager.register(unsplash)
    plugin_manager.register(placeholder)

    # This is the magic that allows our application to discover other plugins
    # installed alongside it
    plugin_manager.load_setuptools_entrypoints(APP_NAME)

    return plugin_manager



from __future__ import annotations

import re
from typing import Any

import click
from pydantic import ValidationError

from ...plugins.manager import (
    SEARCH_BACKEND_SETTINGS_KEY,
    SEARCH_BACKEND_SETTINGS_MODEL,
)

FIELD_SEPARATOR = "."
VALUE_SEPARATOR = ","


class ConfigValuesValidator:
    """
    Validation class that is meant to be called as a validator callable
    within click. It will parse and validate incoming values that look like
    this:

        param=value -> {"param": "value"}
        nested.param=value -> {"nested": {"param": "value"}}
    """

    def __init__(self):
        self._pattern = re.compile(r"([a-zA-Z._]+)=(.*)")

    def __call__(self, ctx, _, values) -> dict[str, str]:
        current_config = ctx.obj.config.dict()
        checked_values = {}

        for value in values:
            checked_values.update(
                self.validate_single_value(value, ctx.obj.config_class)
            )

        try:
            merge(checked_values, current_config)
            ctx.obj.config_class(**current_config)
        except ValidationError as exc:
            raise click.ClickException(f"\n{str(exc)}")

        return current_config

    def validate_single_value(self, value: str, config_class) -> dict:
        """
        Validates a single config parameter. This is done by first matching the
        string value against a regex. Afterwards we return this value in its dictionary
        representation (e.g. "param=value -> {"param": "value"}).

        :raises click.BadParameter: raises to halt program when bad input is received
        """
        match = self._pattern.match(value)

        if not match:
            raise click.BadParameter(self._format_bad_format_error(value))

        parameter, parsed_value = match.groups()
        schema = config_class.schema()
        parameter_type = get_param_type(schema, parameter)

        if parameter_type == tuple:
            parsed_value = parsed_value.split(VALUE_SEPARATOR)

        return get_nested_dict_from_path(parameter, parsed_value)

    @staticmethod
    def _format_bad_format_error(value):
        return (
            f"'{value}' does not conform to the correct format. Please use the 'param=value'"
            " format."
        )


def get_param_type(schema, parameter: str) -> type | None:
    """
    Tries to determine the parameter type based on the schema and the parameter name.

    TODO: This is currently
    """
    first_key, *other_keys = parameter.split(FIELD_SEPARATOR)

    if first_key == SEARCH_BACKEND_SETTINGS_KEY:
        second_key, *other_keys = other_keys
        backend_setting_schema = schema.get("definitions", {}).get(
            SEARCH_BACKEND_SETTINGS_MODEL
        )
        if backend_setting_schema is not None:
            default = (
                backend_setting_schema.get("properties", {})
                .get(second_key, {})
                .get("default")
            )
            if default is not None:
                third_key, *other_keys = other_keys
                return type(default.get(third_key))
    else:
        default = schema.get("properties", {}).get(first_key, {}).get("default")
        if default is not None:
            return type(default)

    raise click.BadParameter(f"'{parameter}' is not a valid configuration parameter")


def get_dotted_path_value(nest: dict, dotted_path: str) -> Any:
    """
    When provides a ``dotted_path`` like "a.b.c" returns the nested value in ``nest``

    Example:
    >>> get_dotted_path_value({"a": {"b": {"c": 1}}}, "a.b.c.")
    1
    >>> get_dotted_path_value({"a": {"b": {"c": 1}}}, "a.b.c.d")  # We go as deep as we can
    1
    >>> get_dotted_path_value({"a": {"b": {"c": 1}}}, "d.c.b.a")

    """
    dotted_path_split = dotted_path.split(FIELD_SEPARATOR)
    current_nest = nest

    for key_val in dotted_path_split:
        sub_val = current_nest.get(key_val)
        if not isinstance(sub_val, dict):
            return sub_val
        else:
            current_nest = sub_val


def get_nested_dict_from_path(path: str, value: Any) -> dict:
    """
    Used to transform our input config parameter "paths" (e.g. "nested.config.param")
    to an expanded dictionary.

    This only supports dictionaries that are structured like dict[str, dict[str, str]].

    Python's recursion limits prohibit deeply nested paths (> 999), but practically,
    this

    Examples:
        >>> get_nested_dict_from_path("nested.config.value", 123)
        {'nested': {'config': {'value': 123}}}
        >>> get_nested_dict_from_path("parameter", 123)
        {'parameter': 123}
        >>> get_nested_dict_from_path("", 123)
        {}
    """
    if not path:
        return {}

    def _inner(_parts: list[str], _value: Any) -> dict:
        if not _parts:
            return _value

        nested_dict = {}
        key = _parts.pop(0)
        nested_dict[key] = _inner(_parts, _value)

        return nested_dict

    parts = path.split(FIELD_SEPARATOR)
    return _inner(parts, value)


def merge(source, destination):
    """
    Recursively merges a dictionary.
    """
    for key, value in source.items():
        if isinstance(value, dict):
            node = destination.setdefault(key, {})
            merge(value, node)
        else:
            destination[key] = value

    return destination
